;----------------------------------------------------------------------
; Fuchsjagdsender mit ATtiny85
;----------------------------------------------------------------------
; Prozessor : ATtiny85
; Takt      : 8 MHz
; Sprache   : Assembler
; Version   : 1.0
; Autor     : Dr. Michael Schramm
; Datum     : 10.2010
;----------------------------------------------------------------------
; Portbelegung: [K1]
; PB0  o: VCC-Versorgung des Senders
; PB1  o: Modulationssignal
; PB2 i : Spannung vom Poti fr Senderhythmus (PB2 = ADC1)
; PB3 io: pull-down ber Widerstand zur Morsesignal-Pegelabsenkung
; PB4  o: VCC-Signal fr LED


; Ergnzende Kommentare sind in kommentar.inc hinterlegt. Aus dieser
; Datei wird mit [K1], [K2] usw. darauf verwiesen.
.include "kommentar.inc"
;----------------------------------------------------------------------
.DSEG ;SRAM-Belegung
m_tonfre: .byte 1
m_geschw: .byte 1
m_anzahl: .byte 1
m_text:   .byte 500
;----------------------------------------------------------------------
.include "tn85def.inc"
.include "makros.inc"
; Einige Register werden als globale Variablen genutzt:
.DEF rtonfre = r7  ;Tonfrequenz-Parameter
.DEF rgeschw = r8  ;Geschwindigkeits-Parameter
.DEF ranzahl = r9  ;Anzahl-Parameter
.DEF mincnt  = r10 ;x..0 Minuten-Countdown fr Senderhythmus
.DEF intreg  = r11 ;nur in Interrupt-Routine genutzt als SREG-Speicher
.DEF tc1mcnt = r12 ;2..-1 T/C1-Minutenzhler fr Korrektur 125/124
.DEF tc1ovfl = r13 ;125..0 T/C1-Overflow-Countdown fr Minutentakt
.DEF minrhy  = r14 ;0..4, Senderhythmus in Minuten laut Potistellung
.DEF nullreg = r15 ;dauerhaft 0
.DEF flags   = r25 ;diverse Flags:
; 0: Volle Minute erreicht (durch T/C-1-Overflow-Interrupt gesetzt)
; 1: T/C-1-Interrupt (durch T/C-1-Overflow-Interrupt gesetzt)
;----------------------------------------------------------------------
.CSEG
;Reset- und Interruptvektoren
  rjmp start   ; 0 RESET, Brown-out Reset, Watchdog Reset
  reti         ; 1 INT0 External Interrupt Request 0
  reti         ; 2 PCINT0 Pin Change Interrupt Request 0
  reti         ; 3 Timer1 CompareA Handler
  rjmp i_tc1ov ; 4 Timer1 Overflow Handler
  reti         ; 5 Timer0 Overflow Handler
  reti         ; 6 EE_RDY EEPROM Ready
  reti         ; 7 ANA_COMP Analog Comparator
  rjmp i_adc   ; 8 ADC ADC Conversion Complete
  reti         ; 9 Timer1 CompareB Handler
  reti         ;10 Timer/Counter0 Compare Match A
  reti ;rjmp i_tc0cmb;11 Timer/Counter0 Compare Match B
; reti         ;12 Watchdog Time-out
; reti         ;13 USI START
; reti         ;14 USI Overflow
;----------------------------------------------------------------------
start: ; zunchst Stromsparmanahmen und Initialisierung
; Port-B-Input-Buffer ausschalten
  out_i DIDR0,(1<<ADC3D)+(1<<ADC2D)+(1<<ADC1D)+(1<<ADC0D)
  out_i DDRB,0b10011 ;Datenrichtung von Port B
  clr nullreg
  out PORTB,nullreg
  sbi ACSR,ACD  ;Analog-Comparator ausschalten
  out_i SPH,high(RAMEND) ;Stackpointer setzen
  out_i SPL,low(RAMEND)
  clr flags
; ADC mit 8bit-Auflsung auf ADC1 = PB2 schalten
  out_i ADMUX,(1<<ADLAR)+(1<<MUX0)

; Den im EEPROM abgelegten zu morsenden Text ins SRAM kopieren
  ldi_hl z,m_tonfre ;Zieladresse im SRAM
  ldi_hl y,0 ;Quelladresse im EEPROM
  ldi r17,4 ;Mindestzahl der zu lesenden Bytes
eerdlp:
  sbic EECR,EEPE ;EEPROM bereit?
  rjmp eerdlp ;nein => warten
  out EEARH,yh ;Leseadresse
  out EEARL,yl
  sbi EECR,EERE ;Read-Signal setzen
  in r0,EEDR ;Datenbyte lesen
  inc r0 ;$FF zu $00 machen
  breq pc+2
  dec r0
  st z+,r0
  adiw yl,1
  dec r17
  brpl eerdlp ;mindestens 4 Bytes lesen
  inc r17
  tst r0 ;$00 ist Endemarke
  brne eerdlp
  ldi_hl z,m_tonfre ;Parameter in Register kopieren
  ld rtonfre,z+
  ld rgeschw,z+
  ld ranzahl,z

; Messung der Potistellung an PB2 zur Festlegung des Senderhythmus:
; Senden alle 0, 1, 2, 3 oder 4 Minuten [K2]
; ADC einschalten mit Interrupt und Vorteiler 64 (=> 125 kHz)
  out_i ADCSRA,(1<<ADEN)+(1<<ADIE)+(6<<ADPS0)
  sei ;Interrupts erlauben
; ADC-Messung durchfhren, Ergebnis in R20
  sleep_adc ;ADC-Noise-Reduction-Modus
  cbi ADCSRA,ADEN ;ADC gleich wieder ausschalten
; Unterscheidung der 5 Bereiche 0..4 auf dem Poti
  ldi r16,-1
  ldi r17,32
potiabfr:
  inc r16
  sub r20,r17
  ldi r17,64
  brcc potiabfr
  mov minrhy,r16 ;R16 muss hier 0..4 enthalten
  mov mincnt,minrhy ;Minuten-Countdown-Zhler

; Timer/Counter 1 als Zeitbasis-Taktgeber [K3]
  out_i OCR1C,234 ;Timer-Reset bei 235 statt 256
; OCR1C als Topwert, Timer-Takt = Systemtakt / 16384
  out_i TCCR1,(1<<CTC1)+(1<<PWM1A)+($F<<CS10)
  ld_i tc1mcnt,2 ;bentigt fr Minutenzhlung durch T/C 1
  ld_i tc1ovfl,125 ;dito
  out_i TIMSK,(1<<TOIE1) ;Overflow-Interrupt ein

blinken:
  rcall blinksignal

; ********************* Start der Sendesequenz *************************
senden:
  sbi PORTB,0 ;Sender einschalten
  sleep_idle ;knapp 0,5 Sekunden warten

audio: ;Audioausgabe per PWM-DAC durch T/C 0 [K4]
  ldi_hl z,2*wavdaten ;hier beginnen die Audiodaten
  ldi_hl x,2*(wavdatend-wavdaten) ;Anzahl der Audio-Bytes
  out_i OCR0B,128 ;Startwert fr DAC
; Timer/Counter 0 als DAC per Fast PWM an Pin PB1 = OC0B
  out_i TCCR0A,(1<<COM0B1)+(3<<WGM00) ;nicht invert. an OC0B
  out_i TCCR0B,(1<<CS00) ;Timertakt = Systemtakt
  out_i TIMSK,(1<<TOIE1)+(1<<OCIE0B) ;CompareMatchB-Interrupt ein
  audiowert:
    lpm r0,z+ ;nchstes Audio-Byte
    out OCR0B,r0 ;Umschaltwert fr PWM
    for r17,4,pwmloop ;4 Timer-0-Durchlufe pro Audiosample
	  sleep_idle
    next_down r17,pwmloop
    sbiw x,1
  brne audiowert
  out TCCR0B,nullreg ;Timer 0 stoppen
  out TCCR0A,nullreg ;Timermodus initialisieren
  rcall wait1_5s ;etwa 1,5 Sekunden warten

  sbi DDRB,3 ;PB3 als Ausgang an GND zur Pegelabsenkung
  mov r20,ranzahl
  morsen: ;Morsezeichen erzeugen
    ldi_hl y,m_text ;hier beginnt der zu morsende Text
    morse_lp:
      ld r16,y+
      tst r16 ;Nullbyte ist Textendemarke
      breq nach_morsen
      rcall morse1z ;R16 als Morsezeichen ausgeben
    rjmp morse_lp
    nach_morsen:
  next_down r20,morsen
  out TCCR0B,nullreg ;Timer 0 stoppen
  out TCCR0A,nullreg ;Timermodus initialisieren
  cbi DDRB,3 ;PB3 wieder als Eingang
  sleep_idle ;in jedem Fall kurz warten mit Sendesignal ohne Modulation
  tst minrhy ;Dauersendung gewnscht?
  brne sender_aus ;nein => dann aus und auf nchsten Einsatz warten

; Im Dauerbetrieb bleibt der Sender permanent eingeschaltet. Nur eine
; kurze Pause ohne Modulation einlegen, dann wieder Audioausgabe.
  rcall wait1_5s ;etwa 1,5 Sekunden warten
  rjmp blinken

sender_aus:
  cbi PORTB,0 ;Sender ausschalten
; ********************* Ende der Sendesequenz *************************

sleepwait: ;warten auf den nchsten Minutenanfang
  sleep_idle
  sbrs flags,0 ;Minuten-Flag gesetzt?
  rjmp sleepwait ;nein => weiter schlafen
  cbr flags,(1<<0) ;Minuten-Flag lschen
  rcall blinksignal ;zu jeder vollen Minute kurzes Blinksignal
  dec mincnt ;Minuten-Countdown-Zhler
  brne sleepwait ;ich bin noch nicht wieder an der Reihe
  mov mincnt,minrhy ;nchste Minutenzhlrunde einluten
  rjmp senden ;wieder auf Sendung gehen

;----------------------------------------------------------------------
; *************************** Unterprogramme **************************

blinksignal: ;kurzes Aufblinken der LED an PB4
; wird immer kurz nach Timer-1-Interrupt aufgerufen, daher wird die LED
; fr knapp 0,5 Sekunden eingeschaltet (bis zum nchsten Interrupt)
  sbi PORTB,4
  sleep_idle
  cbi PORTB,4
ret

wait1_5s: ;etwa 1,5 Sekunden warten
  for r17,3,wait
	sleep_idle
  next_down r17,wait
ret

morse1z: ;ASCII-Zeichen in R16 als Morsezeichen ausgeben [K5]
  andi r16,$7F
  cpi r16,$60
  brcs pc+2
  subi r16,$20 ;Kleinbuchstaben zu Grobuchstaben
  subi r16,$20 ;Steuerzeichen berspringen
  brcc pc+2
  clr r16 ;unzulssige Steuerzeichen durch Space ersetzen
  ldi_hl z,2*morsecode
  add zl,r16
  adc zh,nullreg
  lpm r0,z ;zugehrige Morse-Codierung in R0
  out OCR0A,rtonfre ;Timer zhlt von 0 bis rtonfre
;  mov r1,rtonfre
;  lsr r1
;  out OCR0B,r1 ;bei halbem Wert den Port umschalten
  out_i OCR0B,1 ;bereits bei 1 den Port umschalten
  out TCNT0,nullreg
; Timer/Counter 0 als DAC per Fast PWM an Pin PB1 = OC0B
  out_i TCCR0A,(1<<COM0B1)+(3<<WGM00) ;nicht invert. an OC0B
  out_i TCCR0B,(4<<CS00)+(1<<WGM02) ;Timertakt = Systemtakt / 256 = 31250 Hz
  out_i TIMSK,(1<<TOIE1)+(1<<TOIE0); Overflow-Interrupt ein
; Wortabstandpause erhht von 4 auf 6 dits fr einfacheres Mitlesen
  ldi r17,6 ;zustzliche Pause mit Lnge 6 dits
  tst r0 ;0 steht fr Leerzeichen, also Wortabstandpause
  breq mo_nach_z
mo_next:
  lsr r0 ;nchstes auszugebendes dit/dah
  breq mo_z_ende ;Zeichen zuende, wenn nur noch Nullbits brig
  out_i TCCR0A,(1<<COM0B1)+(3<<WGM00) ;OC0B an PB1 aktivieren
  ldi r17,1 ;dit-Lnge
  brcc mo_signal
  ldi r17,3 ;dah-Lnge
mo_signal: ;dit bzw. dah ausgeben
  rcall mo_r17dit
  ldi r17,1 ;dit-Lnge
  rcall mo_r17dit_ ;Pause zwischen zwei Signalen
  rjmp mo_next

mo_z_ende:
; von 2 auf 3 dits verlngerte Pause zwischen den Zeichen, macht das
; Mithren etwas einfacher (+ 1 dit, das schon gewartet wurde)
  ldi r17,3 ;zustzliche Pause mit Lnge 3 dits nach jedem Zeichen
mo_nach_z:
;  rcall mo_r17dit_
;ret

mo_r17dit_: ;Ton aus, dann wie mo_r17dit
  out_i TCCR0A,(3<<WGM00) ;OC0B an PB1 deaktivieren

mo_r17dit: ;Zeitdauer von r17 dits verstreichen lassen
  mo_paus_lp1:
    for r18,4,mo_paus_lp2
      mov r19,rgeschw
      mo_paus_lp3:
	    cbr flags,(1<<1)
	  mo_paus_lp3b:
		sleep_idle
		sbrc flags,1
		rjmp mo_paus_lp3 ;Interrupt vom "falschen" Timer ignorieren
      next_down r19,mo_paus_lp3b
    next_down r18,mo_paus_lp2
  next_down r17,mo_paus_lp1
ret

;----------------------------------------------------------------------
; ************************* Interrupt-Routinen ************************

i_adc: ;ADC-Ergebnis liegt vor
  in r20,ADCH
reti

i_tc1ov: ; T/C 1 dient als innere Uhr fr den Minutentakt [K3]
  in intreg,SREG
  sbr flags,(1<<1) ;signalisiert T/C-1-Interrupt
  dec tc1ovfl ;125 bzw. 124 Interrupts durch?
  breq itc1_neu ;dann alles fr die nchste Minute setzen
itc1_ret:
  out SREG,intreg ;sonst mglichst schnell zurck
reti
itc1_neu: ;wird einmal pro Minute erreicht
  push r16
  sbr flags,(1<<0) ;signalisiert die volle Minute
  ldi r16,124 ;in einer von 3 Minuten: 124 Interrupts pro Minute
  dec tc1mcnt
  breq itc1_ok1 ;nach der zweiten von 3 Minuten
  ldi r16,125
  brpl itc1_ok1 ;nach der ersten von 3 Minuten
  clr tc1mcnt ; nach der dritten von 3 Minuten ist tc1mcnt = -1
  inc tc1mcnt
  inc tc1mcnt ;3-Minuten-Zhler tc1mcnt auf 2 zurcksetzen
itc1_ok1:
  mov tc1ovfl,r16
  pop r16
rjmp itc1_ret

;----------------------------------------------------------------------
; Codierung der Morsezeichen, ein Byte pro Zeichen. 0 = dit, 1 = dah
; Erstes auszugebendes Signal in Bit 0. Letztes 1-bit markiert Ende.
morsecode: ;fr ASCII-Codes $20 bis $5F
.include "morsecode.inc"

wavdaten: ;Audiodaten 8bit-unsigned, 8000 samples per second
.include "suchMich.inc"
wavdatend:
;----------------------------------------------------------------------
.ESEG ; zu morsender Text im EEPROM
; 1. Byte bestimmt Tonfrequenz
; 2. Byte bestimmt Gebegeschwindigkeit
; 3. Byte ist Wiederholungsfaktor fr nachfolgenden Text
; ab dem 4. Byte steht der Text, Nullbyte als Endemarke
.db 51,20,1,"MOE  AATIS PETLING KLEINER FUCHS",0
; ******************************* ENDE *******************************
